/*
 * ExactPatternMatch.h
 *
 *  Created on: 9. 2. 2014
 *      Author: Jan Travnicek
 */

#ifndef _EXACT_PATTERN_MATCH_H_
#define _EXACT_PATTERN_MATCH_H_

#include <tree/Tree.h>
#include <tree/TreeFeatures.h>
#include <tree/ranked/RankedNode.h>
#include <tree/unranked/UnrankedNode.h>
#include <set>
#include <common/multipleDispatch.hpp>

namespace arbology {

namespace exact {

class ExactPatternMatch : public std::DoubleDispatch < std::set < unsigned >, tree::TreeBase, tree::TreeBase > {
public:
	/**
	 * Performs conversion.
	 * @return left regular grammar equivalent to source automaton.
	 */
	static std::set < unsigned > match ( const tree::Tree & subject, const tree::Tree & pattern );

	static std::set < unsigned > match ( const tree::UnrankedTree & subject, const tree::UnrankedPattern & pattern );
	static std::set < unsigned > match ( const tree::RankedTree & subject, const tree::RankedPattern & pattern );
	static std::set < unsigned > match ( const tree::PrefixRankedTree & subject, const tree::PrefixRankedPattern & pattern );
	static std::set < unsigned > match ( const tree::PrefixRankedBarTree & subject, const tree::PrefixRankedBarPattern & pattern );

private:
	static bool matchHelper ( const tree::UnrankedNode & subject, const tree::UnrankedNode & pattern, const alphabet::Symbol & subtreeVariable );
	static bool matchHelper ( const tree::RankedNode & subject, const tree::RankedNode & pattern, const alphabet::RankedSymbol & subtreeVariable );

	static void matchInternal ( unsigned & index, std::set < unsigned > & occ, const tree::UnrankedNode & subject, const tree::UnrankedNode & pattern, const alphabet::Symbol & subtreeVariable );
	static void matchInternal ( unsigned & index, std::set < unsigned > & occ, const tree::RankedNode & subject, const tree::RankedNode & pattern, const alphabet::RankedSymbol & subtreeVariable );

public:
	static ExactPatternMatch & getInstance ( ) {
		static ExactPatternMatch res;

		return res;
	}

};

} /* namespace exact */

} /* namespace arbology */

#endif /* _EXACT_PATTERN_MATCH_H_ */
