/*
 * ReversedBoyerMooreHorspool.cpp
 *
 *  Created on: 5. 11. 2014
 *      Author: Jan Travnicek
 */

#include "ReversedBoyerMooreHorspool.h"
#include "ReversedBadCharacterShiftTable.h"
#include "SubtreeJumpTable.h"

#include <exception/AlibException.h>
#include <tree/Tree.h>
#include <tree/ranked/PrefixRankedTree.h>
#include <tree/ranked/PrefixRankedBarTree.h>
#include <tree/ranked/PrefixRankedPattern.h>
#include <tree/ranked/PrefixRankedBarPattern.h>
#include <alphabet/RankedSymbol.h>

#include <map>

namespace arbology {

namespace exact {

std::set < unsigned > ReversedBoyerMooreHorspool::match ( const tree::Tree & subject, const tree::Tree & pattern ) {
	return getInstance ( ).dispatch ( subject.getData ( ), pattern.getData ( ) );
}

std::set < unsigned > ReversedBoyerMooreHorspool::match ( const tree::PrefixRankedBarTree & subject, const tree::PrefixRankedBarTree & pattern ) {
	return match ( subject, tree::PrefixRankedBarPattern ( pattern ) );
}

auto ReversedBoyerMooreHorspoolPrefixRankedBarTreePrefixRankedBarTree = ReversedBoyerMooreHorspool::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedBarTree, tree::PrefixRankedBarTree > ( ReversedBoyerMooreHorspool::getInstance ( ), ReversedBoyerMooreHorspool::match );

std::set < unsigned > ReversedBoyerMooreHorspool::match ( const tree::PrefixRankedBarTree & subject, const tree::PrefixRankedBarPattern & pattern ) {
	std::set < unsigned > occ;
	std::map < alphabet::RankedSymbol, size_t > bcs = ReversedBadCharacterShiftTable::bcs ( pattern ); // NOTE: the subjects alphabet must be a subset or equal to the pattern
	std::vector < int > subjectSubtreeJumpTable = SubtreeJumpTable::compute ( subject );

	 // index to the subject
	int i = ( int ) subject.getContent ( ).size ( ) - pattern.getContent ( ).size ( ) + 1;

	 // main loop of the algorithm over all possible indexes where the pattern can start
	while ( i >= 0 ) {

		 // index to the pattern
		unsigned j = 0;

		 // offset to the subject
		unsigned offset = i;

		while ( ( j < pattern.getContent ( ).size ( ) ) && ( offset < subject.getContent ( ).size ( ) ) ) {
			if ( subject.getContent ( )[offset] == pattern.getContent ( )[j] ) {
				 // match of symbol
				offset = offset + 1;
				j = j + 1;
			} else if ( ( pattern.getContent ( )[j] == pattern.getSubtreeWildcard ( ) ) && ( subject.getContent ( )[offset].getSymbol ( ) != pattern.getBarSymbol ( ) ) ) {
				 // match of variable with subtree
				offset = subjectSubtreeJumpTable[offset];
				j = j + 2;
			} else {
				break;
			}
		}

		 // match was found
		if ( j == pattern.getContent ( ).size ( ) ) occ.insert ( i );

		 // shift heristics
		i -= bcs[subject.getContent ( )[i]];
	}

	return occ;
}

auto ReversedBoyerMooreHorspoolPrefixRankedBarTreePrefixRankedBarPattern = ReversedBoyerMooreHorspool::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedBarTree, tree::PrefixRankedBarPattern > ( ReversedBoyerMooreHorspool::getInstance ( ), ReversedBoyerMooreHorspool::match );

std::set < unsigned > ReversedBoyerMooreHorspool::match ( const tree::PrefixRankedTree & subject, const tree::PrefixRankedTree & pattern ) {
	return match ( subject, tree::PrefixRankedPattern ( pattern ) );
}

auto ReversedBoyerMooreHorspoolPrefixRankedBarTreePrefixRankedTree = ReversedBoyerMooreHorspool::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedTree, tree::PrefixRankedTree > ( ReversedBoyerMooreHorspool::getInstance ( ), ReversedBoyerMooreHorspool::match );

std::set < unsigned > ReversedBoyerMooreHorspool::match ( const tree::PrefixRankedTree & subject, const tree::PrefixRankedPattern & pattern ) {
	std::set < unsigned > occ;
	std::map < alphabet::RankedSymbol, size_t > bcs = ReversedBadCharacterShiftTable::bcs ( pattern ); // NOTE: the subjects alphabet must be a subset or equal to the pattern
	std::vector < int > subjectSubtreeJumpTable = SubtreeJumpTable::compute ( subject );

	 // index to the subject
	int i = ( int ) subject.getContent ( ).size ( ) - pattern.getContent ( ).size ( ) + 1;

	 // main loop of the algorithm over all possible indexes where the pattern can start
	while ( i >= 0 ) {

		 // index to the pattern
		unsigned j = 0;

		 // offset to the subject
		unsigned offset = i;

		while ( ( j < pattern.getContent ( ).size ( ) ) && ( offset < subject.getContent ( ).size ( ) ) ) {
			if ( subject.getContent ( )[offset] == pattern.getContent ( )[j] )
				 // match of symbol
				offset = offset + 1;
			else if ( pattern.getContent ( )[j] == pattern.getSubtreeWildcard ( ) )
				 // match of variable with subtree
				offset = subjectSubtreeJumpTable[offset];
			else
				break;

			j = j + 1;
		}

		 // match was found
		if ( j == pattern.getContent ( ).size ( ) ) occ.insert ( i );

		 // shift heristics
		i -= bcs[subject.getContent ( )[i]];
	}

	return occ;
}

auto ReversedBoyerMooreHorspoolPrefixRankedBarTreePrefixRankedPattern = ReversedBoyerMooreHorspool::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedTree, tree::PrefixRankedPattern > ( ReversedBoyerMooreHorspool::getInstance ( ), ReversedBoyerMooreHorspool::match );

} /* namespace exact */

} /* namespace arbology */
